# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import json
import re

from rally.common.io import junit
from rally import consts
from rally.ui import utils as ui_utils
from rally.verification import reporter


SKIP_RE = re.compile(r"Skipped until Bug: ?(?P<bug_number>\d+) is resolved.")
LP_BUG_LINK = "https://launchpad.net/bugs/%s"
TIME_FORMAT = consts.TimeFormat.ISO8601


@reporter.configure("json")
class JSONReporter(reporter.VerificationReporter):
    """Generates verification report in JSON format.

    An example of the report (All dates, numbers, names appearing in this
    example are fictitious. Any resemblance to real things is purely
    coincidental):

      .. code-block:: json

        {"verifications": {
            "verification-uuid-1": {
                "status": "finished",
                "skipped": 1,
                "started_at": "2001-01-01T00:00:00",
                "finished_at": "2001-01-01T00:05:00",
                "tests_duration": 5,
                "run_args": {
                    "pattern": "set=smoke",
                    "xfail_list": {"some.test.TestCase.test_xfail":
                                       "Some reason why it is expected."},
                    "skip_list": {"some.test.TestCase.test_skipped":
                                      "This test was skipped intentionally"},
                },
                "success": 1,
                "expected_failures": 1,
                "tests_count": 3,
                "failures": 0,
                "unexpected_success": 0
            },
            "verification-uuid-2": {
                "status": "finished",
                "skipped": 1,
                "started_at": "2002-01-01T00:00:00",
                "finished_at": "2002-01-01T00:05:00",
                "tests_duration": 5,
                "run_args": {
                    "pattern": "set=smoke",
                    "xfail_list": {"some.test.TestCase.test_xfail":
                                       "Some reason why it is expected."},
                    "skip_list": {"some.test.TestCase.test_skipped":
                                      "This test was skipped intentionally"},
                },
                "success": 1,
                "expected_failures": 1,
                "tests_count": 3,
                "failures": 1,
                "unexpected_success": 0
            }
         },
         "tests": {
            "some.test.TestCase.test_foo[tag1,tag2]": {
                "name": "some.test.TestCase.test_foo",
                "tags": ["tag1","tag2"],
                "by_verification": {
                    "verification-uuid-1": {
                        "status": "success",
                        "duration": "1.111"
                    },
                    "verification-uuid-2": {
                        "status": "success",
                        "duration": "22.222"
                    }
                }
            },
            "some.test.TestCase.test_skipped[tag1]": {
                "name": "some.test.TestCase.test_skipped",
                "tags": ["tag1"],
                "by_verification": {
                    "verification-uuid-1": {
                        "status": "skipped",
                        "duration": "0",
                        "details": "Skipped until Bug: 666 is resolved."
                    },
                    "verification-uuid-2": {
                        "status": "skipped",
                        "duration": "0",
                        "details": "Skipped until Bug: 666 is resolved."
                    }
                }
            },
            "some.test.TestCase.test_xfail": {
                "name": "some.test.TestCase.test_xfail",
                "tags": [],
                "by_verification": {
                    "verification-uuid-1": {
                        "status": "xfail",
                        "duration": "3",
                        "details": "Some reason why it is expected.\\n\\n"
                            "Traceback (most recent call last): \\n"
                            "  File "fake.py", line 13, in <module>\\n"
                            "    yyy()\\n"
                            "  File "fake.py", line 11, in yyy\\n"
                            "    xxx()\\n"
                            "  File "fake.py", line 8, in xxx\\n"
                            "    bar()\\n"
                            "  File "fake.py", line 5, in bar\\n"
                            "    foo()\\n"
                            "  File "fake.py", line 2, in foo\\n"
                            "    raise Exception()\\n"
                            "Exception"
                    },
                    "verification-uuid-2": {
                        "status": "xfail",
                        "duration": "3",
                        "details": "Some reason why it is expected.\\n\\n"
                            "Traceback (most recent call last): \\n"
                            "  File "fake.py", line 13, in <module>\\n"
                            "    yyy()\\n"
                            "  File "fake.py", line 11, in yyy\\n"
                            "    xxx()\\n"
                            "  File "fake.py", line 8, in xxx\\n"
                            "    bar()\\n"
                            "  File "fake.py", line 5, in bar\\n"
                            "    foo()\\n"
                            "  File "fake.py", line 2, in foo\\n"
                            "    raise Exception()\\n"
                            "Exception"
                    }
                }
            },
            "some.test.TestCase.test_failed": {
                "name": "some.test.TestCase.test_failed",
                "tags": [],
                "by_verification": {
                    "verification-uuid-2": {
                        "status": "fail",
                        "duration": "4",
                        "details": "Some reason why it is expected.\\n\\n"
                            "Traceback (most recent call last): \\n"
                            "  File "fake.py", line 13, in <module>\\n"
                            "    yyy()\\n"
                            "  File "fake.py", line 11, in yyy\\n"
                            "    xxx()\\n"
                            "  File "fake.py", line 8, in xxx\\n"
                            "    bar()\\n"
                            "  File "fake.py", line 5, in bar\\n"
                            "    foo()\\n"
                            "  File "fake.py", line 2, in foo\\n"
                            "    raise Exception()\\n"
                            "Exception"
                        }
                    }
                }
            }
        }

    """

    @classmethod
    def validate(cls, output_destination):
        """Validate destination of report.

        :param output_destination: Destination of report
        """
        # nothing to check :)
        pass

    def _generate(self):
        """Prepare raw report."""

        verifications = collections.OrderedDict()
        tests = {}

        for v in self.verifications:
            verifications[v.uuid] = {
                "started_at": v.created_at.strftime(TIME_FORMAT),
                "finished_at": v.updated_at.strftime(TIME_FORMAT),
                "status": v.status,
                "run_args": v.run_args,
                "tests_count": v.tests_count,
                "tests_duration": v.tests_duration,
                "skipped": v.skipped,
                "success": v.success,
                "expected_failures": v.expected_failures,
                "unexpected_success": v.unexpected_success,
                "failures": v.failures,
            }

            for test_id, result in v.tests.items():
                if test_id not in tests:
                    # NOTE(ylobankov): It is more convenient to see test ID
                    #                  at the first place in the report.
                    tags = sorted(result.get("tags", []), reverse=True,
                                  key=lambda tag: tag.startswith("id-"))
                    tests[test_id] = {"tags": tags,
                                      "name": result["name"],
                                      "by_verification": {}}

                tests[test_id]["by_verification"][v.uuid] = {
                    "status": result["status"],
                    "duration": result["duration"]
                }

                reason = result.get("reason", "")
                if reason:
                    match = SKIP_RE.match(reason)
                    if match:
                        link = LP_BUG_LINK % match.group("bug_number")
                        reason = re.sub(match.group("bug_number"), link,
                                        reason)
                traceback = result.get("traceback", "")
                sep = "\n\n" if reason and traceback else ""
                d = (reason + sep + traceback.strip()) or None
                if d:
                    tests[test_id]["by_verification"][v.uuid]["details"] = d

        return {"verifications": verifications, "tests": tests}

    def generate(self):
        raw_report = json.dumps(self._generate(), indent=4)

        if self.output_destination:
            return {"files": {self.output_destination: raw_report},
                    "open": self.output_destination}
        else:
            return {"print": raw_report}


@reporter.configure("html")
class HTMLReporter(JSONReporter):
    """Generates verification report in HTML format."""
    INCLUDE_LIBS = False

    # "T" separator of ISO 8601 is not user-friendly enough.
    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    def generate(self):
        report = self._generate()
        uuids = report["verifications"].keys()
        show_comparison_note = False

        for test in report["tests"].values():
            # make as much as possible processing here to reduce processing
            # at JS side
            test["has_details"] = False
            for test_info in test["by_verification"].values():
                if "details" not in test_info:
                    test_info["details"] = None
                elif not test["has_details"]:
                    test["has_details"] = True

            durations = []
            # iter by uuids to store right order for comparison
            for uuid in uuids:
                if uuid in test["by_verification"]:
                    durations.append(test["by_verification"][uuid]["duration"])
                    if float(durations[-1]) < 0.001:
                        durations[-1] = "0"
                        # not to display such little duration in the report
                        test["by_verification"][uuid]["duration"] = ""

                    if len(durations) > 1 and not (
                            durations[0] == "0" and durations[-1] == "0"):
                        # compare result with result of the first verification
                        diff = float(durations[-1]) - float(durations[0])
                        result = "%s (" % durations[-1]
                        if diff >= 0:
                            result += "+"
                        result += "%s)" % diff
                        test["by_verification"][uuid]["duration"] = result

            if not show_comparison_note and len(durations) > 2:
                # NOTE(andreykurilin): only in case of comparison of more
                #   than 2 results of the same test we should display a note
                #   about the comparison strategy
                show_comparison_note = True

        template = ui_utils.get_template("verification/report.html")
        context = {"uuids": list(uuids),
                   "verifications": report["verifications"],
                   "tests": report["tests"],
                   "show_comparison_note": show_comparison_note}

        raw_report = template.render(data=json.dumps(context),
                                     include_libs=self.INCLUDE_LIBS)

        # in future we will support html_static and will need to save more
        # files
        if self.output_destination:
            return {"files": {self.output_destination: raw_report},
                    "open": self.output_destination}
        else:
            return {"print": raw_report}


@reporter.configure("html-static")
class HTMLStaticReporter(HTMLReporter):
    """Generates verification report in HTML format with embedded JS/CSS."""
    INCLUDE_LIBS = True


@reporter.configure("junit-xml")
class JUnitXMLReporter(reporter.VerificationReporter):
    """Generates verification report in JUnit-XML format.

    An example of the report (All dates, numbers, names appearing in this
    example are fictitious. Any resemblance to real things is purely
    coincidental):

      .. code-block:: xml

        <testsuites>
          <!--Report is generated by Rally 0.8.0 at 2002-01-01T00:00:00-->
          <testsuite id="verification-uuid-1"
                     tests="9"
                     time="1.111"
                     errors="0"
                     failures="3"
                     skipped="0"
                     timestamp="2001-01-01T00:00:00">
            <testcase classname="some.test.TestCase"
                      name="test_foo"
                      time="8"
                      timestamp="2001-01-01T00:01:00" />
            <testcase classname="some.test.TestCase"
                      name="test_skipped"
                      time="0"
                      timestamp="2001-01-01T00:02:00">
              <skipped>Skipped until Bug: 666 is resolved.</skipped>
            </testcase>
            <testcase classname="some.test.TestCase"
                      name="test_xfail"
                      time="3"
                      timestamp="2001-01-01T00:03:00">
              <!--It is an expected failure due to: something-->
              <!--Traceback:
        HEEELP-->
            </testcase>
            <testcase classname="some.test.TestCase"
                      name="test_uxsuccess"
                      time="3"
                      timestamp="2001-01-01T00:04:00">
              <failure>
                  It is an unexpected success. The test should fail due to:
                  It should fail, I said!
              </failure>
            </testcase>
          </testsuite>
          <testsuite id="verification-uuid-2"
                     tests="99"
                     time="22.222"
                     errors="0"
                     failures="33"
                     skipped="0"
                     timestamp="2002-01-01T00:00:00">
            <testcase classname="some.test.TestCase"
                      name="test_foo"
                      time="8"
                      timestamp="2001-02-01T00:01:00" />
            <testcase classname="some.test.TestCase"
                      name="test_failed"
                      time="8"
                      timestamp="2001-02-01T00:02:00">
              <failure>HEEEEEEELP</failure>
            </testcase>
            <testcase classname="some.test.TestCase"
                      name="test_skipped"
                      time="0"
                      timestamp="2001-02-01T00:03:00">
              <skipped>Skipped until Bug: 666 is resolved.</skipped>
            </testcase>
            <testcase classname="some.test.TestCase"
                      name="test_xfail"
                      time="4"
                      timestamp="2001-02-01T00:04:00">
              <!--It is an expected failure due to: something-->
              <!--Traceback:
        HEEELP-->
            </testcase>
          </testsuite>
        </testsuites>

    """

    @classmethod
    def validate(cls, output_destination):
        pass

    def generate(self):
        report = junit.JUnitXML()

        for v in self.verifications:
            test_suite = report.add_test_suite(
                id=v.uuid,
                time=str(v.tests_duration),
                timestamp=v.created_at.strftime(TIME_FORMAT)
            )
            test_suite.setup_final_stats(
                tests=str(v.tests_count),
                skipped=str(v.skipped),
                failures=str(v.failures + v.unexpected_success)
            )

            tests = sorted(v.tests.values(),
                           key=lambda t: (t.get("timestamp", ""), t["name"]))
            for result in tests:
                class_name, name = result["name"].rsplit(".", 1)

                test_id = [tag[3:] for tag in result.get("tags", [])
                           if tag.startswith("id-")]

                test_case = test_suite.add_test_case(
                    id=(test_id[0] if test_id else None),
                    time=result["duration"], name=name, classname=class_name,
                    timestamp=result.get("timestamp"))

                if result["status"] == "success":
                    # nothing to add
                    pass
                elif result["status"] == "uxsuccess":
                    test_case.mark_as_uxsuccess(
                        result.get("reason"))
                elif result["status"] == "fail":
                    test_case.mark_as_failed(
                        result.get("traceback", None))
                elif result["status"] == "xfail":
                    trace = result.get("traceback", None)
                    test_case.mark_as_xfail(
                        result.get("reason", None),
                        f"Traceback:\n{trace}" if trace else None)
                elif result["status"] == "skip":
                    test_case.mark_as_skipped(
                        result.get("reason", None))
                else:
                    # wtf is it?! we should add validation of results...
                    pass

        raw_report = report.to_string()
        if self.output_destination:
            return {"files": {self.output_destination: raw_report},
                    "open": self.output_destination}
        else:
            return {"print": raw_report}



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def get_tasks_results():
    task_id = "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"
    workload = {"uuid": "uuid",
                "name": "CinderVolumes.list_volumes",
                "description": "List all volumes.",
                "created_at": "2017-06-04T05:14:44",
                "updated_at": "2017-06-04T05:15:14",
                "task_uuid": task_id,
                "position": 0,
                "data": [
                    {
                        "duration": 0.2504892349243164,
                        "timestamp": 1584551892.7336202,
                        "idle_duration": 0.0,
                        "error": [],
                        "output": {
                            "additive": [],
                            "complete": []
                        },
                        "atomic_actions": [
                            {
                                "name": "foo",
                                "children": [
                                    {
                                        "name": "bar",
                                        "children": [],
                                        "started_at": 1584551892.733688,
                                        "finished_at": 1584551892.984079
                                    }
                                ],
                                "started_at": 1584551892.7336745,
                                "finished_at": 1584551892.9840994
                            }
                        ]
                    },
                    {
                        "duration": 0.25043749809265137,
                        "timestamp": 1584551892.7363858,
                        "idle_duration": 0.0,
                        "error": [],
                        "output": {
                            "additive": [],
                            "complete": []
                        },
                        "atomic_actions": [
                            {
                                "name": "foo",
                                "children": [
                                    {
                                        "name": "bar",
                                        "children": [],
                                        "started_at": 1584551892.7364488,
                                        "finished_at": 1584551892.9867969
                                    }
                                ],
                                "started_at": 1584551892.736435,
                                "finished_at": 1584551892.9868152
                            }
                        ]
                    }
                ],
                "full_duration": 29.969523191452026,
                "load_duration": 2.03029203414917,
                "hooks": [],
                "runner": {},
                "runner_type": "runner_type",
                "args": {},
                "contexts": {},
                "contexts_results": [],
                "min_duration": 0.0,
                "max_duration": 1.0,
                "start_time": 0,
                "statistics": {},
                "failed_iteration_count": 0,
                "total_iteration_count": 10,
                "sla": {},
                "sla_results": {"sla": []},
                "pass_sla": True
                }
    task = {
        "uuid": task_id,
        "title": "task",
        "description": "description",
        "status": "finished",
        "env_uuid": "env-uuid",
        "env_name": "env-name",
        "tags": [],
        "created_at": "2017-06-04T05:14:44",
        "updated_at": "2017-06-04T05:15:14",
        "pass_sla": True,
        "task_duration": 2.0,
        "subtasks": [
            {"uuid": "subtask_uuid",
             "title": "subtask",
             "description": "description",
             "status": "finished",
             "run_in_parallel": True,
             "created_at": "2017-06-04T05:14:44",
             "updated_at": "2017-06-04T05:15:14",
             "sla": {},
             "duration": 29.969523191452026,
             "task_uuid": task_id,
             "workloads": [workload]}
        ]}
    return [task]



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for HookExecutor and Hook classes."""

from unittest import mock

import ddt

from rally import consts
from rally.task import hook
from tests.unit import fakes
from tests.unit import test


@hook.configure(name="dummy_hook")
class DummyHookAction(hook.HookAction):
    CONFIG_SCHEMA = {
        "type": "object",
        "properties": {
            "status": {"type": "string"},
            "error": {"type": "object"},
            "output": {"type": "object"},
        },
        "required": ["status"],
        "additionalProperties": False,
    }

    def run(self):
        self.set_status(self.config["status"])

        error = self.config.get("error")
        if error:
            self.set_error(*error)

        output = self.config.get("output")
        if output:
            self.add_output(**output)


class HookExecutorTestCase(test.TestCase):

    def setUp(self):
        super(HookExecutorTestCase, self).setUp()
        self.conf = {
            "hooks": [
                {
                    "description": "dummy_action",
                    "action": (
                        "dummy_hook", {"status": consts.HookStatus.SUCCESS}
                    ),
                    "trigger": (
                        "event", {
                            "unit": "iteration",
                            "at": [1],
                        }
                    )
                }
            ]
        }
        self.task = mock.MagicMock()

    @mock.patch("rally.task.hook.HookExecutor._timer_method")
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_results(self, mock_timer, mock__timer_method):
        hook_executor = hook.HookExecutor(self.conf, self.task)
        hook_executor.on_event(event_type="iteration", value=1)

        self.assertEqual(
            [{"config": self.conf["hooks"][0],
              "results": [{
                  "triggered_by": {"event_type": "iteration", "value": 1},
                  "started_at": fakes.FakeTimer().timestamp(),
                  "finished_at": fakes.FakeTimer().finish_timestamp(),
                  "status": consts.HookStatus.SUCCESS}],
              "summary": {consts.HookStatus.SUCCESS: 1}}],
            hook_executor.results())

    @mock.patch("rally.task.hook.HookExecutor._timer_method")
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_result_optional(self, mock_timer, mock__timer_method):
        hook_args = self.conf["hooks"][0]["action"][1]
        hook_args["error"] = ["Exception", "Description", "Traceback"]

        hook_args["output"] = {"additive": None, "complete": None}

        hook_executor = hook.HookExecutor(self.conf, self.task)
        hook_executor.on_event(event_type="iteration", value=1)

        self.assertEqual(
            [{"config": self.conf["hooks"][0],
              "results": [{
                  "triggered_by": {"event_type": "iteration", "value": 1},
                  "started_at": fakes.FakeTimer().timestamp(),
                  "finished_at": fakes.FakeTimer().finish_timestamp(),
                  "error": {"details": "Traceback", "etype": "Exception",
                            "msg": "Description"},
                  "output": {"additive": [], "complete": []},
                  "status": consts.HookStatus.FAILED}],
              "summary": {consts.HookStatus.FAILED: 1}}],
            hook_executor.results())

    def test_empty_result(self):
        hook_executor = hook.HookExecutor(self.conf, self.task)
        self.assertEqual([{"config": self.conf["hooks"][0],
                           "results": [],
                           "summary": {}}],
                         hook_executor.results())

    @mock.patch("rally.task.hook.HookExecutor._timer_method")
    @mock.patch.object(DummyHookAction, "run",
                       side_effect=Exception("My err msg"))
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_failed_result(self, mock_timer, mock_dummy_hook_action_run,
                           mock__timer_method):
        hook_executor = hook.HookExecutor(self.conf, self.task)
        hook_executor.on_event(event_type="iteration", value=1)

        self.assertEqual(
            [{"config": self.conf["hooks"][0],
              "results": [{
                  "triggered_by": {"event_type": "iteration", "value": 1},
                  "error": {"etype": "Exception",
                            "msg": mock.ANY, "details": mock.ANY},
                  "started_at": fakes.FakeTimer().timestamp(),
                  "finished_at": fakes.FakeTimer().finish_timestamp(),
                  "status": consts.HookStatus.FAILED}],
              "summary": {consts.HookStatus.FAILED: 1}}],
            hook_executor.results())

    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_time_event(self, mock_timer):
        trigger_args = self.conf["hooks"][0]["trigger"][1]
        trigger_args["unit"] = "time"

        hook_executor = hook.HookExecutor(self.conf, self.task)
        hook_executor.on_event(event_type="time", value=1)

        self.assertEqual(
            [{"config": self.conf["hooks"][0],
              "results": [{
                  "triggered_by": {"event_type": "time", "value": 1},
                  "started_at": fakes.FakeTimer().timestamp(),
                  "finished_at": fakes.FakeTimer().finish_timestamp(),
                  "status": consts.HookStatus.SUCCESS}],
              "summary": {consts.HookStatus.SUCCESS: 1}}],
            hook_executor.results())

    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_time_periodic(self, mock_timer):
        self.conf["hooks"][0]["trigger"] = ("periodic",
                                            {"unit": "time", "step": 2})
        hook_executor = hook.HookExecutor(self.conf, self.task)

        for i in range(1, 7):
            hook_executor.on_event(event_type="time", value=i)

        self.assertEqual(
            [{
                "config": self.conf["hooks"][0],
                "results": [
                    {
                        "triggered_by": {"event_type": "time", "value": 2},
                        "started_at": fakes.FakeTimer().timestamp(),
                        "finished_at": fakes.FakeTimer().finish_timestamp(),
                        "status": consts.HookStatus.SUCCESS
                    },
                    {
                        "triggered_by": {"event_type": "time", "value": 4},
                        "started_at": fakes.FakeTimer().timestamp(),
                        "finished_at": fakes.FakeTimer().finish_timestamp(),
                        "status": consts.HookStatus.SUCCESS
                    },
                    {
                        "triggered_by": {"event_type": "time", "value": 6},
                        "started_at": fakes.FakeTimer().timestamp(),
                        "finished_at": fakes.FakeTimer().finish_timestamp(),
                        "status": consts.HookStatus.SUCCESS
                    }
                ],
                "summary": {consts.HookStatus.SUCCESS: 3}
            }],
            hook_executor.results())

    @mock.patch("rally.common.utils.Stopwatch", autospec=True)
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_timer_thread(self, mock_timer, mock_stopwatch):
        trigger_args = self.conf["hooks"][0]["trigger"][1]
        trigger_args["unit"] = "time"
        hook_executor = hook.HookExecutor(self.conf, self.task)

        def stop_timer(sec):
            if sec == 3:
                hook_executor._timer_stop_event.set()

        stopwatch_inst = mock_stopwatch.return_value
        stopwatch_inst.sleep.side_effect = stop_timer

        hook_executor.on_event(event_type="iteration", value=1)
        self.assertTrue(hook_executor._timer_stop_event.wait(1))

        self.assertEqual(
            [{"config": self.conf["hooks"][0],
              "results": [{
                  "triggered_by": {"event_type": "time", "value": 1},
                  "started_at": fakes.FakeTimer().timestamp(),
                  "finished_at": fakes.FakeTimer().finish_timestamp(),
                  "status": consts.HookStatus.SUCCESS}],
              "summary": {consts.HookStatus.SUCCESS: 1}
              }],
            hook_executor.results())

        stopwatch_inst.start.assert_called_once_with()
        stopwatch_inst.sleep.assert_has_calls([
            mock.call(1),
            mock.call(2),
            mock.call(3),
        ])


@ddt.ddt
class HookTestCase(test.TestCase):

    @ddt.data(({"status": "foo"}, True), (3, False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = hook.HookAction.validate(
            "dummy_hook", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    def test_result(self, mock_timer):
        task = mock.MagicMock()
        triggered_by = {"event_type": "iteration", "value": 1}
        dummy_hook = DummyHookAction(
            task, {"status": consts.HookStatus.SUCCESS}, triggered_by)
        dummy_hook.run_sync()

        self.assertEqual(
            {"started_at": fakes.FakeTimer().timestamp(),
             "finished_at": fakes.FakeTimer().finish_timestamp(),
             "triggered_by": triggered_by,
             "status": consts.HookStatus.SUCCESS}, dummy_hook.result())

    def test_result_not_started(self):
        task = mock.MagicMock()
        triggered_by = {"event_type": "iteration", "value": 1}
        dummy_hook = DummyHookAction(task,
                                     {"status": consts.HookStatus.SUCCESS},
                                     triggered_by)

        self.assertEqual(
            {"started_at": 0.0,
             "finished_at": 0.0,
             "triggered_by": triggered_by,
             "status": consts.HookStatus.SUCCESS}, dummy_hook.result())


@ddt.ddt
class TriggerTestCase(test.TestCase):

    def setUp(self):
        super(TriggerTestCase, self).setUp()

        @hook.configure(name="dummy_trigger")
        class DummyTrigger(hook.HookTrigger):
            CONFIG_SCHEMA = {"type": "array",
                             "minItems": 1,
                             "uniqueItems": True,
                             "items": {
                                 "type": "integer",
                                 "minimum": 0,
                             }}

            def get_listening_event(self):
                return "dummy"

            def on_event(self, event_type, value=None):
                if value not in self.config:
                    return
                super(DummyTrigger, self).on_event(event_type, value)

        self.DummyTrigger = DummyTrigger

        self.addCleanup(DummyTrigger.unregister)

    @ddt.data(([5], True), ("str", False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = hook.HookTrigger.validate(
            "dummy_trigger", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    def test_on_event_and_get_results(self):

        # get_results requires launched hooks, so if we want to test it, we
        # need to duplicate all calls on_event. It is redundant, so let's merge
        # test_on_event and test_get_results in one test.
        right_values = [5, 7, 12, 13]

        cfg = {"trigger": (self.DummyTrigger.get_name(), right_values),
               "action": ("fake", {})}
        task = mock.MagicMock()
        hook_cls = mock.MagicMock(__name__="fake")
        dummy_trigger = self.DummyTrigger(cfg, task, hook_cls)
        for i in range(0, 20):
            dummy_trigger.on_event("fake", i)

        self.assertEqual(
            [mock.call(task, {}, {"event_type": "fake", "value": i})
             for i in right_values],
            hook_cls.call_args_list)
        self.assertEqual(len(right_values),
                         hook_cls.return_value.run_async.call_count)
        hook_status = hook_cls.return_value.result.return_value["status"]
        res = [hook_cls.return_value.result.return_value] * len(right_values)
        self.assertEqual(
            {"config": cfg,
             "results": res,
             "summary": {hook_status: len(right_values)}},
            dummy_trigger.get_results())



#!/usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import subprocess
import sys
import uuid


def create_cfg_for_using_sqlite():
    """Create a config file for rally with path to temporary sqlite database"""
    run_id = str(uuid.uuid4())[:8]

    cfg_path = f"/tmp/self_rally_{run_id}.conf"

    cfg = (f"[database]\n"
           f"connection = sqlite:////tmp/self-rally-{run_id}.sqlite")
    with open(cfg_path, "w") as f:
        f.write(cfg)

    return "--config-file", cfg_path


def _print_with_wrapper(*messages):
    print("#" * 80)
    for m in messages:
        print(f"# {m}")
    print("#" * 80)


class Rally(object):
    def __init__(self, rally_cmd):
        self._rally_cmd = rally_cmd

    def __call__(self, cmd, debug=False, description=None):
        if not isinstance(cmd, (tuple, list)):
            cmd = cmd.split(" ")

        final_cmd = self._rally_cmd[:]
        if debug:
            final_cmd.append("--debug")
        final_cmd.extend(cmd)

        messages = [f"Calling: {' '.join(final_cmd)}"]
        if description:
            messages.append(f"Description: {description}")
        _print_with_wrapper(*messages)

        p = subprocess.Popen(
            final_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        for line in iter(p.stdout.readline, b""):
            print(line.rstrip().decode("utf-8"))
        p.wait()
        if p.returncode != 0:
            _print_with_wrapper(
                f"Command returned non-zero exit status {p.returncode}.")
            sys.exit(1)


def main():
    parser = argparse.ArgumentParser(
        "rally_self_job",
        description="A script that ensures that Rally works after installation"
    )
    parser.add_argument(
        "--results-dir", type=str, default=".test_results",
        help="A path to directory to place rally results execution."
    )
    parser.add_argument(
        "--task", type=str, required=True,
        help="A simple task to launch."
    )
    parser.add_argument(
        "--without-tmp-sqlite", action="store_false",
        dest="prepare_sqlite",
        help="Do not create a tmp sqlite db for Rally"
    )
    parser.add_argument(
        "--rally-cmd", type=str, default="rally",
        help="An override 'rally' entry point for all commands."
    )
    parser.add_argument(
        "--plugins-path", type=str, required=True,
        help="A path to external rally plugins that include "
             "'FakePlugin.testplugin'"
    )

    args = parser.parse_args()

    rally_cmd = args.rally_cmd.split(" ")
    if "--config-file" in rally_cmd:
        raise Exception(
            "It is restricted to use '--config-file' at rally cmd.")

    if args.prepare_sqlite:
        rally_cmd.extend(create_cfg_for_using_sqlite())

    rally_cmd.append(f"--plugin-paths={args.plugins_path}")

    rally = Rally(rally_cmd)

    rally("--version")
    rally("plugin show --name FakePlugin.testplugin", debug=True,
          description="Ensure plugins loading")

    rally("db create", description="Initialize temporary sqlite database")

    rally("env create --name=self",
          description="Create empty environment")
    rally(f"task start {args.task}")

    results_dir = args.results_dir
    rally(f"task report --html-static --out {results_dir}/self_report.html")
    rally(f"task report --json --out {results_dir}/self_report.json")
    rally("task sla-check")


if __name__ == "__main__":
    main()
